import os
import sys
import operator
from abc import ABC, abstractmethod
from typing import Dict, Any, TypeVar, Generic, Iterable, Union, Tuple
import math
from enum import Enum, unique
from collections import defaultdict, namedtuple
import torch
import torch.nn as nn
from torch.utils.data import DataLoader
from torch.optim.optimizer import Optimizer
from torch.optim.lr_scheduler import _LRScheduler
from torchland.datasets.loader_maker import DataLoaderBuilder
from torch.utils.tensorboard import SummaryWriter


"""Tuple storing model information."""
if sys.hexversion >= 0x3070000:  # 'defaults' keyword appeared in ver. 3.7
    ModelInfo = namedtuple(
        'ModelInfo', ['model', 'input_size', 'metric', 'comparison'], defaults=[operator.lt])
else:
    ModelInfo = namedtuple('ModelInfo', ['model', 'input_size', 'metric', 'comparison'])
    ModelInfo.__new__.__defaults__ = (operator.lt, )


@unique
class TrainStage(Enum):
    """Enum defining each stages of training process."""
    TRAIN = 'train'
    VALIDATE = 'validate'
    TEST = 'test'


class MetricManager:
    """Class for managing multiple metrics."""
    def __init__(self):
        self.metric_counter = defaultdict(int)  # init to 0
        self.metric_avgs = defaultdict(float)  # init to 0.0
        self.metric_mins = defaultdict(lambda: math.inf)
        self.metric_maxes = defaultdict(lambda: -math.inf)

    def append_metric(self, metric: dict):
        """
        Introduce a new metric values and update the statistics.
        It mainly updates the count, average value, minimum value, and the maximum value.

        Args:
            metric (dict): various metric values
        """
        for key, val in metric.items():
            prev_count = self.metric_counter[key]
            prev_avg = self.metric_avgs[key]
            total_val = prev_count * prev_avg + val

            # calculate the new average
            self.metric_avgs[key] = total_val / (prev_count + 1)
            self.metric_counter[key] = prev_count + 1
            if val < self.metric_mins[key]:
                self.metric_mins[key] = val
            if val > self.metric_maxes[key]:
                self.metric_maxes[key] = val

    def mean(self, key: str) -> float:
        """
        Retrieve the mean value of the given key.

        Args:
            key (str): the key value

        Returns:
            the mean value of the key
        """
        return self.metric_avgs[key]

    def set_mean(self, key: str, val):
        self.metric_avgs[key] = val


T = TypeVar('T')


class AttributeHolder(Generic[T], Iterable):
    def __init__(self):
        self.num_attrs = 0

    def add(self, name: str, attr: Any):
        setattr(self, name, attr)
        self.num_attrs += 1

    def empty(self):
        return self.num_attrs == 0

    def __len__(self):
        return self.num_attrs

    def __iter__(self):
        return iter(filter(lambda s: s != 'num_attrs', vars(self)))

    def __getitem__(self, item):
        return getattr(self, item)


class NetworkTrainer(ABC):
    """
    Generalized Neural Network trainer that make it easy to customize for different
    models, dataloaders, criterions, etc.
    It also provides basic logging, model saving, and summary writing.
    ``fit()`` method provides a standardized training process - train-validate-test.
    """

    def __init__(self,
                 epoch: int,
                 output_dir='data_out',
                 num_devices=1,
                 seed: int=None,
                 lr_scheduler: Dict[str, _LRScheduler]=None,
                 log_every_local=50,
                 save_module_every_local=500):
        """
        Initialize the trainer.

        Args:
            epoch (int): total epochs to train (the end epoch)
            output_dir (str): root output directory
            num_devices (int): number of GPU devices to split the batch
            seed (int): random seed to use
            lr_scheduler (None|Dict[str, _LRScheduler]): learning rate scheduler
            log_every_local (int): log the progress every `log_every_local` steps
            save_module_every_local (int): saves module information per this amount of steps
        """
        # initial settings
        self.device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
        # training devices to use
        self.device_ids = list(range(num_devices))

        # set seed
        if seed is None:
            self.seed = torch.initial_seed()
        else:
            self.seed = seed
            torch.manual_seed(seed)
        print(f'Using random seed : {self.seed}')

        self.trainer_name = self.__class__.__name__

        # prepare model(s) for training
        self.models: AttributeHolder[ModelInfo] = AttributeHolder()
        self.criteria: AttributeHolder[nn.Module] = AttributeHolder()
        self.optimizers: AttributeHolder[Optimizer] = AttributeHolder()

        # dataloaders - should be set using set_dataloader_builder() after __init__()
        self.train_dataloader = None
        self.val_dataloader = None
        self.test_dataloader = None

        # setup and create output directories
        self.output_dir = output_dir
        self.log_dir = self._create_output_dir('logs')
        self.model_dir = self._create_output_dir('models')
        self.onnx_dir = self._create_output_dir('onnx')
        self.checkpoint_dir = self._create_output_dir('checkpoints')

        # save any other states or variables to maintain
        self.total_epoch = epoch
        self.lr_schedulers = lr_scheduler
        self.log_every_local = log_every_local
        self.save_module_every_local = save_module_every_local
        self.writer = SummaryWriter(log_dir=self.log_dir)

        # initialize training process
        self.epoch = 0
        self.global_step = 0  # total steps run
        self.local_step = 0  # local step within a single epoch

    def add_model(self, name: str, model: nn.Module, input_size: Tuple[int, ...], metric: str):
        self.models.add(
            name,
            ModelInfo(
                model=self._register_model(model),
                input_size=input_size,
                metric=metric))

    def add_optimizer(self, name: str, optimizer: Optimizer):
        self.optimizers.add(name, optimizer)

    def add_criterion(self, name: str, criteria: nn.Module):
        self.criteria.add(name, criteria)

    def set_dataloader_builder(self, dataloader_builder: DataLoaderBuilder):
        if dataloader_builder is None:
            raise ValueError('dataloader builder should not be None')

        self.train_dataloader = dataloader_builder.make_train_dataloader()
        self.val_dataloader = dataloader_builder.make_validate_dataloader()
        self.test_dataloader = dataloader_builder.make_test_dataloader()

    @property
    def standard_metric(self):
        """Define the standard metric to compare the performance of models."""
        return 'loss'

    def _create_output_dir(self, dirname: str):
        """Creates a directory in the root output directory."""
        path = os.path.join(self.output_dir, dirname)
        os.makedirs(path, exist_ok=True)
        return path

    def fit(self, use_val_metric=True):
        """Run the entire training process."""
        print(f'Using validation metric for best model : {use_val_metric}')
        best_metric = None
        self.pre_fit()

        for _ in range(self.epoch, self.total_epoch):
            print(f'################# Starting epoch {self.epoch} ##################')

            train_metrics = self.train()

            # run upon validation set
            val_metrics = self.validate()

            # compare the metric and save the best model
            target_metric = val_metrics if use_val_metric else train_metrics
            best_metric = self._save_best_model(self.models, best_metric, target_metric)

            # update learning rate based on validation metric
            self._update_lr(val_metrics)

            self.epoch += 1
        # run upon test set
        test_metrics = self.test()
        print('Training complete.')

    def pre_fit(self):
        """
        Anything that could be run before the fitting starts.
        """
        pass

    def _update_lr(self, val_metrics):
        if self.lr_schedulers is not None:
            for lrs in self.lr_schedulers.values():
                lrs.step()

    @abstractmethod
    def run_step(
            self,
            models: AttributeHolder[ModelInfo],
            criteria: AttributeHolder[nn.Module],
            optimizers: AttributeHolder[Optimizer],
            input_: Union[torch.Tensor, Tuple[torch.Tensor]],
            train_stage: TrainStage,
            *args, **kwargs):
        """
        Run a single step.
        It is given all required instances for training.

        Args:
            models (AttributeHolder[ModelInfo]): models to train
            criteria (AttributeHolder[nn.Module]): model criteria functions
            optimizers (AttributeHolder[Optimizer]): model optimizers
            input_ (Union[torch.Tensor, Tuple[torch.Tensor]]): inputs to models
            train_stage (TrainStage): enum indicating which stage of training it is going through

        Returns:
            (output, loss) : any combination of outputs and loss values
        """
        raise NotImplementedError

    def _run_epoch(self, dataloader: DataLoader, train_stage: TrainStage):
        """
        Runs an epoch for a dataset.

        Args:
            dataloader (DataLoader): data loader to use in this epoch
            train_stage (TrainStage): enum indicating which training step this epoch is running

        Returns:
            metric_manager: a set of important metrics managed by the manager
        """
        self.before_epoch(train_stage=train_stage)

        metric_manager = MetricManager()  # initialize the metric manager
        dataset_size = len(dataloader)
        for step, input_ in enumerate(dataloader):
            self.local_step = step
            input_ = self._to_device(self.input_transform(input_))  # transform dataloader's data

            # run a single step
            # this step is exposed to public for custom implementation
            # the parameters passed should have equal form as passed into the constructor
            output, loss = self.run_step(
                self.models,
                self.criteria,
                self.optimizers,
                input_,
                train_stage)

            # metric calculation
            metric = self.make_performance_metric(input_, output, loss)
            metric_manager.append_metric(metric)

            # perform any action required after running the step
            self.post_step(
                input_, output, metric, dataset_size=dataset_size, train_stage=train_stage)

            # update train step if training
            if train_stage == TrainStage.TRAIN:
                self.global_step += 1

            # run pre-epoch-finish after the final step
            if step == dataset_size - 1:
                self.pre_epoch_finish(input_, output, metric_manager, train_stage=train_stage)
        self.on_epoch_finish(
            metric_manager, dataset_size=dataset_size, train_stage=train_stage)
        return metric_manager

    def test(self):
        with torch.no_grad():
            results = self._run_epoch(self.test_dataloader, TrainStage.TEST)
        return results

    def train(self):
        # train (model update)
        return self._run_epoch(self.train_dataloader, TrainStage.TRAIN)

    def validate(self):
        with torch.no_grad():
            results = self._run_epoch(self.val_dataloader, TrainStage.VALIDATE)
        return results

    def _register_model(self, model: nn.Module):
        return torch.nn.parallel.DataParallel(
            model.to(self.device), device_ids=self.device_ids)

    @staticmethod
    def make_performance_metric(input_: torch.Tensor, output, loss) -> dict:
        return {'loss': loss.item()}

    def before_epoch(self, train_stage: TrainStage):
        # store the epoch number (to look good in tensorboard), and learning rate
        if train_stage == TrainStage.TRAIN:
            self.writer.add_scalar('epoch', self.epoch, self.global_step)
            self.save_learning_rate(self.writer, self.optimizers, self.global_step)

    def post_step(
            self, input, output, metric: dict,
            dataset_size: int, train_stage: TrainStage):
        if train_stage == TrainStage.TRAIN:
            if self.local_step % self.log_every_local == 0:
                self.log_metric(
                    self.writer, metric, self.epoch, self.global_step,
                    self.local_step, dataset_size, train_stage.value)

            # save model information
            if self.local_step % self.save_module_every_local == 0:
                self._save_module_summary_all()

    def pre_epoch_finish(
            self,
            input_: torch.Tensor,
            output: tuple,
            metric_manager: MetricManager,
            train_stage: TrainStage):
        pass

    def on_epoch_finish(
            self, metric_manager: MetricManager, dataset_size: int, train_stage: TrainStage):
        if train_stage == TrainStage.VALIDATE or train_stage == TrainStage.TEST:
            self.log_metric(
                self.writer, metric_manager.metric_avgs, self.epoch,
                self.global_step, self.local_step,
                dataset_size=dataset_size,
                summary_group=train_stage.value)

        if train_stage == TrainStage.TRAIN:
            self.save_checkpoint()

        # save all modules at the end of epoch
        self._save_all_modules()

    def save_checkpoint(self, prefix=''):
        """Saves the model and training checkpoint.
        The model only saves the model, and it is usually used for inference in the future.
        The checkpoint saves the state dictionary of various modules
        required for training. Usually this information is used to resume training.
        """
        # get state_dict for all models
        model_state = {}
        for model_name in self.models:
            model_state[model_name] = self.models[model_name].model.state_dict()

        optimizer_state = {}
        for optim_name in self.optimizers:
            optimizer_state[optim_name] = self.optimizers[optim_name].state_dict()

        train_state = {
            'epoch': self.epoch,
            'step': self.global_step,
            'seed': self.seed,
            'models': model_state,  # tuple of states (even for len == 1)
            'optimizers': optimizer_state,  # tuple of states
        }
        cptf = f'{prefix}{self.trainer_name}_checkpoint_e{self.epoch}.pth'
        torch.save(train_state, os.path.join(self.checkpoint_dir, cptf))

    def resume(self, filename: str):
        """
        Load checkpoint file and set internal fields accordingly.

        Args:
            filename (str): file path to checkpoint file
        """
        cpt = torch.load(filename)
        self.seed = cpt['seed']
        torch.manual_seed(self.seed)
        self.epoch = cpt['epoch']
        self.global_step = cpt['step']

        # load the model and optimizer
        model_state = cpt['models']
        for model_name in self.models:
            model_info = self.models[model_name]
            model = model_info.model
            # replace ModelInfo's field
            self.models[model_name] = model_info._replace(
                model=model.load_state_dict(model_state[model_name]))

        self.optimizers = [o.load_state_dict(state_dict)
                           for o, state_dict in zip(self.optimizers, cpt['optimizers'])]

    def cleanup(self):
        self.writer.close()

    @staticmethod
    def input_transform(data):
        """Provide an adapter between dataloader outputs and model inputs.
        It is an id() function by default.
        """
        return data

    def _save_best_model(self, models: AttributeHolder[ModelInfo], prev_best_metric, curr_metric):
        if prev_best_metric is None:
            return curr_metric

        best_metric = prev_best_metric
        for model_name in models:
            model, input_size, compare_metric, comparison = models[model_name]
            # compare the standard metric, and if the standard performance metric
            # is better, then save the best model
            if comparison(
                    curr_metric.mean(compare_metric),
                    prev_best_metric.mean(compare_metric)):
                # onnx model saving may fail due to unsupported operators, etc.
                try:
                    self._save_module(model.module, input_size, save_onnx=True, prefix='best_')
                except ImportError as onnx_err:
                    print(f'Failed to import package onnx: {onnx_err}')
                except RuntimeError as onnx_err:
                    print(f'Saving onnx model failed : {onnx_err}')
                self._save_module(model.module, input_size, prefix='best')
                best_metric.set_mean(compare_metric, curr_metric.mean(compare_metric))
        return best_metric

    def _save_all_modules(self):
        for model_name in self.models:
            model_info = self.models[model_name]
            self._save_module(
                model_info.model.module, model_info.input_size)

    def _save_module(self, module, input_size: tuple, save_onnx=False, prefix=''):
        """
        Saves a single module.

        Args:
            module (nn.Module): module to be saved
            input_size (tuple): input dimensions
            save_onnx (bool): save in ONNX format if True
            prefix: prefix for file name
        """
        if save_onnx:
            import onnx
            # TODO: input / output names?
            path = os.path.join(self.onnx_dir, f'{prefix}{module.__class__.__name__}_onnx.pth')
            # add batch dimension to the dummy input sizes
            dummy_input = torch.randn((1, ) + input_size).to(self.device)
            torch.onnx.export(module, dummy_input, path, verbose=True)
            # check validity of onnx IR and print the graph
            model = onnx.load(path)
            onnx.checker.check_model(model)
            onnx.helper.printable_graph(model.graph)
        else:
            # note epoch for default prefix
            if prefix == '':
                prefix = f'e{self.epoch:03}'
            path = os.path.join(self.model_dir, f'{prefix}_{module.__class__.__name__}.pth')
            torch.save(module, path)

    @staticmethod
    def log_metric(
            writer, metrics: dict, epoch: int,
            global_step: int, local_step: int, dataset_size: int, summary_group='train'):
        log = f'Epoch ({summary_group}): {epoch:03} ({local_step}/{dataset_size}) global_step: {global_step}\t'
        for metric_name, val in metrics.items():
            log += f'{metric_name}: {val:.06f}    '
            # write to summary writer
            writer.add_scalar(f'{summary_group}/{metric_name}', val, global_step)
        print(log)

    @staticmethod
    def save_learning_rate(
            writer, optimizers, step: int):
        for opt_name in optimizers:
            opt = optimizers[opt_name]
            for idx, param_group in enumerate(opt.param_groups):
                lr = param_group['lr']
                writer.add_scalar(f'lr/{opt_name}', lr, step)
                print(f'Learning rate for optimizer {opt_name}: {lr}')

    def _save_module_summary_all(self, **kwargs):
        for model_name in self.models:
            model_info = self.models[model_name]
            self._save_module_summary(
                self.writer, model_name, model_info.model.module, self.global_step, **kwargs)

    @staticmethod
    def _save_module_summary(
            writer, module_name: str, module: nn.Module, step: int,
            save_histogram=False, verbose=False):
        # warning: saving histograms is expensive - both time and space
        if module_name == '' or module_name is None:
            module_name = module.__class__.__name__  # to distinguish among different modules

        with torch.no_grad():
            for p_name, parameter in module.named_parameters():
                if parameter.grad is not None:
                    avg_grad = torch.mean(parameter.grad)
                    if verbose:
                        print(f'\tavg_grad for {p_name}_{module_name} = {avg_grad:.6f}')
                    writer.add_scalar(f'avg_grad/{module_name}_{p_name}', avg_grad.item(), step)
                    if save_histogram:
                        writer.add_histogram(
                            f'grad/{module_name}_{p_name}', parameter.grad.cpu().numpy(), step)

                if parameter.data is not None:
                    avg_weight = torch.mean(parameter.data)
                    if verbose:
                        print(f'\tavg_weight for {module_name}_{p_name} = {avg_weight:.6f}')
                    writer.add_scalar(f'avg_weight/{module_name}_{p_name}', avg_weight.item(), step)
                    if save_histogram:
                        writer.add_histogram(
                            f'weight/{module_name}_{p_name}', parameter.data.cpu().numpy(), step)

    def _to_device(self, data):
        """
        Send the data to device this trainer is using

        Args:
            data (tuple|list|torch.Tensor): data

        Returns:
            device-transferred data
        """
        if isinstance(data, tuple):
            return tuple([d.to(self.device) for d in data])
        elif isinstance(data, list):
            return [d.to(self.device) for d in data]
        return data.to(self.device)



import os
import math
import random
import gzip
import urllib.request
from itertools import islice
import numpy as np
from PIL import Image
from torch.utils.data import Dataset, DataLoader, sampler
import torchvision.transforms as transforms
from torchland.utils.noise import zero_mask_noise


TRAIN_IMG_URL = 'http://yann.lecun.com/exdb/mnist/train-images-idx3-ubyte.gz'
TRAIN_LABEL_URL = 'http://yann.lecun.com/exdb/mnist/train-labels-idx1-ubyte.gz'
TEST_IMG_URL = 'http://yann.lecun.com/exdb/mnist/t10k-images-idx3-ubyte.gz'
TEST_LABEL_URL = 'http://yann.lecun.com/exdb/mnist/t10k-labels-idx1-ubyte.gz'


def to_displayable_form(img_tensor):
    return img_tensor.numpy().astype(np.uint8).resize((28, 28))


def load_noisy_mnist_dataloader(batch_size: int, img_shape: tuple=None):
    """Creates dataloaders with this noisy MNIST dataset."""
    data_root = './sdae/mnist'
    train_dataset = NoisyMnistDataset(data_root, img_shape, train=True, zero_prob=0.40)
    # train=True because validation set is split from training dataset
    validate_dataset = NoisyMnistDataset(data_root, img_shape, train=True, zero_prob=0.40)
    test_dataset = NoisyMnistDataset(data_root, img_shape, train=False, zero_prob=0.50)

    num_data = len(train_dataset)
    indices = list(range(num_data))
    random.shuffle(indices)
    num_train = math.floor(num_data * 0.9)
    train_idx, validate_idx = indices[:num_train], indices[num_train:]
    train_dataloader = DataLoader(
        train_dataset,
        sampler=sampler.SubsetRandomSampler(train_idx),
        pin_memory=True,
        drop_last=True,
        num_workers=4,
        batch_size=batch_size,
    )
    validate_dataloader = DataLoader(
        validate_dataset,
        sampler=sampler.SubsetRandomSampler(validate_idx),
        pin_memory=True,
        drop_last=True,
        num_workers=4,
        batch_size=batch_size,
    )
    test_dataloader = DataLoader(
        test_dataset,
        pin_memory=True,
        drop_last=True,
        num_workers=4,
        shuffle=True,
        batch_size=batch_size,
    )
    return train_dataloader, validate_dataloader, test_dataloader


class NoisyMnistDataset(Dataset):
    """MNIST dataset that produces pairs of images: the original 28 by 28 image
    and the same image corrupted by setting pixels to 0 with probability `zero_prob`.

    This is used to train a denoising autoencoder, so no class label is required.
    """
    def __init__(self, data_root: str, img_shape: tuple=None, train=True, zero_prob=0.25):
        super().__init__()
        self.zero_prob = zero_prob
        if img_shape is not None:
            self.img_shape = img_shape
        else:
            self.img_shape = (1, 28, 28)
        self.image_size = 784

        # create data input path
        os.makedirs(data_root, exist_ok=True)

        if train:
            img_path = os.path.join(data_root, 'train_img.gz')
            label_path = os.path.join(data_root, 'train_label.gz')
            img_url = TRAIN_IMG_URL
            label_url = TRAIN_LABEL_URL
        else:
            img_path = os.path.join(data_root, 'test_img.gz')
            label_path = os.path.join(data_root, 'test_label.gz')
            img_url = TEST_IMG_URL
            label_url = TEST_LABEL_URL

        # load image data
        if not os.path.exists(img_path):
            print('Downloading train images: {}'.format(img_path))
            train_data_response = urllib.request.urlopen(img_url)
            with open(img_path, 'wb') as tif:
                tif.write(train_data_response.read())

        with gzip.open(img_path, 'rb') as tif:
            train_data_bytes = tif.read()[16:]
        byte_iterator = map(int, train_data_bytes)
        # slice by 784 elements and accumulate as list until it sees a blank array ([])
        self.train_data = list(iter(lambda: list(islice(byte_iterator, self.image_size)),
                                    []))

        # load label data
        if not os.path.exists(label_path):
            print('Downloading train labels')
            train_label_response = urllib.request.urlopen(label_url)
            with open(label_path, 'wb') as tlf:
                tlf.write(train_label_response.read())

        with gzip.open(label_path, 'rb') as tlf:
            train_label_bytes = tlf.read()[8:]
        self.train_label = list(map(int, train_label_bytes))

    def __getitem__(self, idx):
        img = np.asarray(self.train_data[idx], dtype=np.uint8)
        # corrupt with zero mask noise
        img_corrupted = zero_mask_noise(img, zero_prob=self.zero_prob, dtype=np.uint8)

        img = img.reshape((28, 28))
        img = Image.fromarray(img, mode='L')  # mode 'L' = the array contains 0 - 255 integer values

        img_corrupted = img_corrupted.reshape((28, 28))
        img_corrupted = Image.fromarray(img_corrupted, mode='L')

        transform = transforms.Compose([
            transforms.ToTensor(),
            transforms.Normalize((0.1307, ), (0.3081, )),
        ])
        return transform(img).view(*self.img_shape), transform(img_corrupted).view(*self.img_shape)

    def __len__(self):
        return len(self.train_data)


if __name__ == '__main__':
    dataset = NoisyMnistDataset(data_root='./sdae/mnist')
    c, o = dataset[1]
    print(c.shape)



from torch.utils.data import Dataset, DataLoader


class RandomDataset(Dataset):
    def __init__(self, shape, size: int):
        super().__init__()

    def __getitem__(self, item):
        pass

    def __len__(self):
        pass



"""
Provides method of creating dataloaders of commonly used datasets.
All methods should commonly return a tuple of 4:
    1. train dataloader
    2. validation dataloader
    3. test dataloader
    4. any miscellaneous information / data about the dataset in dict form
        (None if ... well, none)
"""
import os
import math
import random
from torch.utils import data
import torchvision.datasets as datasets
import torchvision.transforms as transforms
from .loader_maker import DataLoaderBuilder


class MNISTLoaderBuilder(DataLoaderBuilder):
    """Data loader maker for MNIST dataset."""
    def __init__(self, data_root: str, batch_size: int, num_workers=4, naive_normalization=False):
        super().__init__()
        self.batch_size = batch_size
        self.num_classes = 10
        self.num_workers = num_workers
        self.dim = 28

        train_img_dir = os.path.join(data_root, 'mnist')

        # use different normalizer by option - the latter one uses the sample distribution's statistics
        # Be CAREFUL of sample distribution based normalization!!
        # the values will not be within boundaries of [-1, 1]
        normalizer = transforms.Normalize(mean=(0.5, ), std=(0.5, )) \
            if naive_normalization else transforms.Normalize((0.1307, ), (0.3081, ))

        # image transform (normalization)
        mnist_transform = transforms.Compose([
            transforms.ToTensor(),
            normalizer,
        ])

        # create datasets
        self.train_dataset = datasets.MNIST(
            root=train_img_dir,
            train=True, download=True,
            transform=mnist_transform)
        self.validate_dataset = datasets.MNIST(
            root=train_img_dir,
            train=True, download=True,
            transform=mnist_transform)
        # create test dataset
        self.test_dataset = datasets.MNIST(
            root=train_img_dir,
            train=False, download=True,
            transform=mnist_transform)

        # split indices btwn. train and validation sets
        num_data = len(self.train_dataset)
        indices = list(range(num_data))
        random.shuffle(indices)
        num_train = math.floor(num_data * 0.9)
        self.train_idx, self.validate_idx = \
            indices[:num_train], indices[num_train:]

    def make_train_dataloader(self):
        train_dataloader = data.DataLoader(
            self.train_dataset,
            sampler=data.sampler.SubsetRandomSampler(self.train_idx),
            pin_memory=True,
            drop_last=True,
            num_workers=self.num_workers,
            batch_size=self.batch_size,
        )
        return train_dataloader

    def make_validate_dataloader(self):
        validate_dataloader = data.DataLoader(
            self.validate_dataset,
            sampler=data.sampler.SubsetRandomSampler(self.validate_idx),
            pin_memory=True,
            drop_last=True,
            num_workers=self.num_workers,
            batch_size=self.batch_size,
        )
        return validate_dataloader

    def make_test_dataloader(self):
        test_dataloader = data.DataLoader(
            self.test_dataset,
            pin_memory=True,
            drop_last=True,
            num_workers=self.num_workers,
            shuffle=True,
            batch_size=self.batch_size,
        )
        return test_dataloader


class LSUNLoaderBuilder(DataLoaderBuilder):
    """DataLoader maker for LSUN (Large-scale scene understanding) dataset."""
    def __init__(self, data_root: str, batch_size: int, num_workers=4):
        super().__init__()
        self.batch_size = batch_size
        self.num_workers = num_workers
        self.img_dim = 64  # default to 64 x 64

        img_dir = os.path.join(data_root, 'lsun')

        # image transform (normalization)
        lsun_transform = transforms.Compose([
            transforms.Resize(self.img_dim),
            transforms.CenterCrop(self.img_dim),
            transforms.ToTensor(),
            transforms.Normalize(mean=(0.5, 0.5, 0.5), std=(0.5, 0.5, 0.5))
        ])

        # create datasets
        self.train_dataset = datasets.LSUN(
            root=img_dir,
            classes=['bedroom_train'],  # TODO: allow other classes
            transform=lsun_transform)
        self.validate_dataset = datasets.LSUN(
            root=img_dir,
            classes=['bedroom_val'],
            transform=lsun_transform)
        # create test dataset - this will share the training dataset
        self.test_dataset = datasets.LSUN(
            root=img_dir,
            classes=['bedroom_train'],
            transform=lsun_transform)

        # split indices btwn. train and validation sets
        num_data = len(self.train_dataset)
        indices = list(range(num_data))
        random.shuffle(indices)
        num_train = math.floor(num_data * 0.95)
        self.train_idx, self.test_idx = \
            indices[:num_train], indices[num_train:]

    def make_train_dataloader(self):
        return data.DataLoader(
            self.train_dataset,
            sampler=data.sampler.SubsetRandomSampler(self.train_idx),
            pin_memory=True,
            drop_last=True,
            num_workers=self.num_workers,
            batch_size=self.batch_size,
        )

    def make_validate_dataloader(self):
        return data.DataLoader(
            self.validate_dataset,
            pin_memory=True,
            drop_last=True,
            shuffle=True,
            num_workers=self.num_workers,
            batch_size=self.batch_size,
        )

    def make_test_dataloader(self):
        return data.DataLoader(
            self.test_dataset,
            sampler=data.sampler.SubsetRandomSampler(self.test_idx),
            pin_memory=True,
            drop_last=True,
            num_workers=self.num_workers,
            batch_size=self.batch_size,
        )


class ImageNetLoaderBuilder(DataLoaderBuilder):
    """
    Create a DataLoader maker for ImageNet dataset.
    This dataset is commonly used for ILSVRC (Imagenet Large Scale Visual Recognition Challenge) competition,
    and has been used in popularity since the AlexNet's success.

    The dataset commonly contains more than 1,000,000 images of 1000 classes.

    See Also:
        http://www.image-net.org/
        http://www.image-net.org/challenges/LSVRC/
    """
    def __init__(self, data_root: str, batch_size: int, num_workers: int,
                 img_dim=224, naive_normalization=False):
        super().__init__()
        self.num_classes = 1000
        self.img_dim = img_dim
        self.data_root = data_root
        self.batch_size = batch_size
        self.num_workers = num_workers

        # image transform sequence for train and validation sets
        if naive_normalization:
            normalize = transforms.Normalize(
                mean=[0.5, 0.5, 0.5], std=[0.5, 0.5, 0.5])
        else:
            normalize = transforms.Normalize(
                mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225])

        self.train_transform = transforms.Compose([
            # randomly resize the image
            # transforms.RandomResizedCrop(self.img_dim, scale=(0.9, 1.0), ratio=(0.9, 1.1)),
            # transforms.RandomHorizontalFlip(),
            transforms.Resize(self.img_dim),
            transforms.CenterCrop(self.img_dim),
            transforms.ToTensor(),
            normalize,
        ])
        self.test_transform = transforms.Compose([
            transforms.Resize(self.img_dim),  # do not adjust the aspect ratio, just resize the image
            transforms.CenterCrop(self.img_dim),
            transforms.ToTensor(),
            normalize,
        ])

        # create datasets
        self.train_dataset = datasets.ImageFolder(self.data_root, self.train_transform)
        self.validate_dataset = datasets.ImageFolder(self.data_root, self.train_transform)
        self.test_dataset = datasets.ImageFolder(self.data_root, self.test_transform)

        # split the indices
        num_data = len(self.train_dataset)
        indices = list(range(num_data))
        random.shuffle(indices)
        # split train <-> (val + test)
        num_train = math.floor(num_data * 0.8)
        self.train_indices, valtest_indices = indices[:num_train], indices[num_train:]
        # split val <-> test
        num_validate = math.floor(num_data * 0.1)
        self.validate_indices = valtest_indices[:num_validate]
        self.test_indices = valtest_indices[num_validate:]

    def make_train_dataloader(self) -> data.DataLoader:
        # create data loaders out of datasets
        return data.DataLoader(
            self.train_dataset,
            sampler=data.sampler.SubsetRandomSampler(self.train_indices),
            pin_memory=True,
            drop_last=True,
            num_workers=self.num_workers,
            batch_size=self.batch_size,
        )

    def make_validate_dataloader(self) -> data.DataLoader:
        return data.DataLoader(
            self.validate_dataset,
            sampler=data.sampler.SubsetRandomSampler(self.validate_indices),
            pin_memory=True,
            drop_last=True,
            num_workers=self.num_workers,
            batch_size=self.batch_size,
        )

    def make_test_dataloader(self) -> data.DataLoader:
        return data.DataLoader(
            self.test_dataset,
            sampler=data.sampler.SubsetRandomSampler(self.test_indices),
            pin_memory=True,
            drop_last=True,
            num_workers=self.num_workers,
            batch_size=self.batch_size,
        )


class CIFAR10LoaderBuilder(DataLoaderBuilder):
    """DataLoader maker for CIFAR-10 dataset.

    See Also: https://www.cs.toronto.edu/~kriz/cifar.html
    """
    DEFAULT_IMG_DIM = 32

    def __init__(self, data_root: str, batch_size: int, num_workers=4,
                 img_dim=32, naive_normalization=False):
        super().__init__()
        self.num_workers = num_workers
        self.batch_size = batch_size
        self.num_classes = 10
        self.img_dim = img_dim

        # data normalizers
        if naive_normalization:
            normalizer = transforms.Normalize(mean=[0.5, 0.5, 0.5], std=[0.5, 0.5, 0.5])
        else:
            normalizer = transforms.Normalize(mean=[0.4914, 0.4822, 0.4465], std=[0.2023, 0.1994, 0.2010])

        train_img_dir = os.path.join(data_root, 'cifar10')

        resizer = []
        # resize the image if it does not match the original
        if self.img_dim != self.DEFAULT_IMG_DIM:
            resizer.append(transforms.Resize(self.img_dim))

        trainval_transform = transforms.Compose(resizer + [
            transforms.RandomCrop(img_dim, padding=4),
            transforms.RandomHorizontalFlip(),
            transforms.ToTensor(),
            normalizer,
        ])

        test_transform = transforms.Compose(resizer + [
            transforms.CenterCrop(size=img_dim),
            transforms.ToTensor(),
            normalizer,
        ])

        self.train_dataset = datasets.CIFAR10(
            root=train_img_dir,
            train=True, download=True,
            transform=trainval_transform)
        self.validate_dataset = datasets.CIFAR10(
            root=train_img_dir,
            train=True, download=True,
            transform=trainval_transform)
        self.test_dataset = datasets.CIFAR10(
            root=train_img_dir,
            train=False, download=True,
            transform=test_transform)

        # split indices btwn. train and validation sets
        num_data = len(self.train_dataset)
        indices = list(range(num_data))
        random.shuffle(indices)
        num_train = math.floor(num_data * 0.9)
        self.train_idx, self.validate_idx = indices[:num_train], indices[num_train:]

    def make_train_dataloader(self):
        return data.DataLoader(
            self.train_dataset,
            sampler=data.sampler.SubsetRandomSampler(self.train_idx),
            pin_memory=True,
            drop_last=True,
            num_workers=self.num_workers,
            batch_size=self.batch_size,
        )

    def make_validate_dataloader(self):
        return data.DataLoader(
            self.validate_dataset,
            sampler=data.sampler.SubsetRandomSampler(self.validate_idx),
            pin_memory=True,
            drop_last=True,
            num_workers=self.num_workers,
            batch_size=self.batch_size,
        )

    def make_test_dataloader(self):
        return data.DataLoader(
            self.test_dataset,
            shuffle=True,
            pin_memory=True,
            drop_last=True,
            num_workers=self.num_workers,
            batch_size=self.batch_size,
        )



"""
Provides methods for creating data loaders of image transfer tasks.
"""
import os
import math
import random
import itertools
import torch
from torch.utils import data
from torchvision.datasets import folder
import torchvision.transforms as transforms
from .loader_maker import DataLoaderBuilder


class Monet2PhotoDataset(data.Dataset):
    """
    Dataset containing pairs of Monet paintings and photographs.
    See CycleGAN paper: https://arxiv.org/abs/1703.10593,
    and dataset dowload script: https://github.com/junyanz/pytorch-CycleGAN-and-pix2pix/blob/master/datasets/download_cyclegan_dataset.sh
    for reference.
    """
    def __init__(self, monet_path: str, photo_path: str, transform=None, shuffle=True, random_samp: int=200000):
        """
        Args:
            monet_path(str): path to images of Monet paintings
            photo_path(str): path to photographs
            transform(callable): image transform chains
            shuffle(bool): if True, shuffles the ordering of pairs
            random_samp(int): if not None, randomly sample this amount from the entire set
        """
        super().__init__()
        self.monets = [os.path.join(monet_path, fname) for fname in os.listdir(monet_path)]
        self.photos = [os.path.join(photo_path, fname) for fname in os.listdir(photo_path)]
        self.img_loader = folder.default_loader  # default image loader provided by torchvision
        self.transform = transform

        if shuffle:
            random.shuffle(self.monets)
            random.shuffle(self.photos)

        # create all possible pairs of paths of images
        self.datapairs = list(itertools.product(self.monets, self.photos))
        if random_samp is not None and random_samp <= len(self.datapairs):
            self.datapairs = random.sample(self.datapairs, random_samp)

    def __getitem__(self, index):
        monet_path, photo_path = self.datapairs[index]

        # load images
        monet_img = self.img_loader(monet_path)
        if self.transform is not None:
            monet_img = self.transform(monet_img)

        photo_img = self.img_loader(photo_path)
        if self.transform is not None:
            photo_img = self.transform(photo_img)

        # return a pair of images - will have size: [2, ch, h, w]
        return torch.stack((monet_img, photo_img))

    def __len__(self):
        return len(self.datapairs)


class Monet2PhotoLoaderBuilder(DataLoaderBuilder):
    """
    DataLoader maker for Monet-to-Photo dataset, renown for being used by CycleGAN.
    """
    def __init__(self, batch_size: int, root_dir='./monet2photo',
                 train_monet_dir='trainA', train_photo_dir='trainB',
                 test_monet_dir='testA', test_photo_dir='testB',
                 random_samp=200000, downsize_half=False, num_workers=8):
        super().__init__()
        self.batch_size = batch_size
        self.root_dir = root_dir
        self.num_workers = num_workers

        # collect path information
        self.train_monet_path = os.path.join(root_dir, train_monet_dir)
        self.train_photo_path = os.path.join(root_dir, train_photo_dir)
        self.test_monet_path = os.path.join(root_dir, test_monet_dir)
        self.test_photo_path = os.path.join(root_dir, test_photo_dir)

        # check that all path exists
        for path in [
            self.train_monet_path,
            self.train_photo_path,
            self.train_monet_path,
            self.test_photo_path,
        ]:
            if not os.path.exists(path):
                print(f'The path: {path} does not exist!')
                raise FileNotFoundError(path)

        # downsize 256 x 256 image into half sized images
        self.img_dim = 256
        if downsize_half:
            self.img_dim = 128
        monet2photo_transform = transforms.Compose([
            transforms.Resize(size=(int(self.img_dim * 1.2), int(self.img_dim * 1.2))),
            transforms.RandomCrop(size=(self.img_dim, self.img_dim)),
            transforms.RandomHorizontalFlip(),
            transforms.ToTensor(),
            transforms.Normalize(mean=(0.5, 0.5, 0.5), std=(0.5, 0.5, 0.5)),
        ])
        self.img_size = (3, self.img_dim, self.img_dim)

        # create datasets that will be loaded to each dataloader instances
        self.train_dataset = Monet2PhotoDataset(
            monet_path=self.train_monet_path,
            photo_path=self.train_photo_path,
            transform=monet2photo_transform,
            random_samp=random_samp)
        self.test_dataset = Monet2PhotoDataset(
            monet_path=self.test_monet_path,
            photo_path=self.test_photo_path,
            transform=monet2photo_transform,
            random_samp=random_samp)

        num_train_data = len(self.train_dataset)
        indices = list(range(num_train_data))
        random.shuffle(indices)
        # usually a generative model does not require a large validation set
        num_trains = math.floor(num_train_data * 0.95)
        self.train_indices, self.val_indices = \
            indices[:num_trains], indices[num_trains:]

        self.default_dataloader_kwargs = {
            'pin_memory': True,
            'drop_last': True,
            'num_workers': self.num_workers,
            'batch_size': self.batch_size,
        }

    def make_train_dataloader(self):
        return data.DataLoader(
            self.train_dataset,
            sampler=data.sampler.SubsetRandomSampler(self.train_indices),
            **self.default_dataloader_kwargs)

    def make_validate_dataloader(self):
        return data.DataLoader(
            self.train_dataset,
            sampler=data.sampler.SubsetRandomSampler(self.val_indices),
            **self.default_dataloader_kwargs)

    def make_test_dataloader(self):
        return data.DataLoader(
            self.test_dataset,
            shuffle=True,
            **self.default_dataloader_kwargs)


if __name__ == '__main__':
    transform = transforms.Compose([
        transforms.ToTensor(),
        transforms.Normalize(mean=(0.5, 0.5, 0.5), std=(0.5, 0.5, 0.5)),
    ])

    monet2photo_dataset = Monet2PhotoDataset(
        monet_path='./cyclegan/datasets/monet2photo/trainA',
        photo_path='./cyclegan/datasets/monet2photo/trainB',
        transform=transform)

    monet2photo_loadermaker = Monet2PhotoLoaderBuilder(
        batch_size=10, root_dir='./cyclegan/datasets/monet2photo/', downsize_half=True)
    train_dataloader = monet2photo_loadermaker.make_train_dataloader()

    # print the size and length of the dataset
    print(len(train_dataloader))
    for img_pair in train_dataloader:
        print(img_pair.size())
        print(img_pair[:, 1, :, :].size())
        break



import librosa
import numpy as np
import scipy


def split_spectrogram(spec, chunk_size_in_frames: int):
    """
    Splits spectrogram into chunks that have sizes of
    'chunk_size' in the time axis.

    Args:
        spec (np.ndarray): spectrogram represented as 2D array
        chunk_size_in_frames (int): number of frames to split into

    Returns:
        split_spec = list of np.ndarray's representing chunks having ``chunk_size`` sizes
    """
    time_length = spec.shape[1]
    num_chunks = time_length // chunk_size_in_frames
    spec = spec[:, :num_chunks * chunk_size_in_frames]
    return np.hsplit(spec, num_chunks)


def denormalize_db_spectrogram(spec, high=20.0, low=-100.0):
    """
    Denormalizes normalized spectrum into db-spectrum.

    Args:
        spec: normalized spectrum
        high (float): maximum db level
        low (float): mean db level

    Returns:
        denormalized db spectrum
    """
    mid = (low + high) / 2.0  # -40
    scale = (high - mid)  # 60
    return (spec * scale) + mid


def normalize_db_spectrogram(db_spec, high=20.0, low=-100.0):
    """
    Normalizes db spectrum with zero mean and unit variance.
    In normal cases, librosa generates db spectrum having -100dB as maximum.

    Args:
        db_spec: db spectrum
        high (float): maximum decibel value
        low (float): minimum decibel value

    Returns:
        normalized db spectrum
    """
    mid = (low + high) / 2.0  # -40
    scale = (high - mid)  # 60
    return (db_spec - mid) / scale


def recover_spectrogram(power_spec, original_phase):
    """
    Recover spectrogram from power spectrogram.
    In order to sufficiently recover the original spectrogram,
    phase information is required.

    Args:
        power_spec (np.ndarray): power spectrogram
        original_phase (np.ndarray): phase information
            - usually extracted from original spectorgram

    Returns:
        recovered_spec (np.ndarray): recovered spectrogram
    """
    assert power_spec.shape == original_phase.shape
    # take the abs since power_spec may have negative values
    abs_power_spec = np.abs(power_spec)
    return np.multiply(np.sqrt(abs_power_spec), np.exp(1j * original_phase))


def recover_audio_from_mel_spectrogram(mel_spec, original_spec,
                                       n_fft: int, fs: int, n_mels: int):
    """
    Recover and create audio from mel-spectrogram.
    In order to sufficiently recover the audio, original spectrogram as well
    as some information about the original audio is required.

    Args:
        mel_spec (np.ndarray): mel-spectrogram
        original_spec (np.ndarray): original spectrogram
        n_fft (int): number of FFT bins
        fs (int): sampling rate of original audio
        n_mels (int): number of mel-spectrogram bins

    Returns:
        ys (np.ndarray): recovered audio signal
    """
    mel_basis = librosa.filters.mel(sr=fs, n_fft=n_fft, n_mels=n_mels)
    # power spectrum recovered
    recovered_power_spec = np.dot(np.matrix(mel_basis).I, mel_spec)
    recovered_spec = recover_spectrogram(recovered_power_spec, np.angle(original_spec))
    # return reconstructed audio
    _, ys = scipy.signal.istft(recovered_spec, fs=fs, n_fft=n_fft)
    return ys



import numpy as np


def zero_mask_noise(x, zero_prob=0.25, dtype=None):
    """
    Randomly 'erase' the data by setting them to 0
    with bernoulli probability `zero_prob`.
    """
    if dtype is None:
        dtype = x.dtype
    zero_mask = np.random.choice([0, 1], size=x.shape, p=[zero_prob, 1 - zero_prob])
    return (x * zero_mask).astype(dtype)



from abc import ABC, abstractmethod
from torch.utils.data import DataLoader


class DataLoaderBuilder(ABC):
    def __init__(self):
        pass

    @abstractmethod
    def make_train_dataloader(self) -> DataLoader:
        raise NotImplementedError

    @abstractmethod
    def make_test_dataloader(self) -> DataLoader:
        raise NotImplementedError

    @abstractmethod
    def make_validate_dataloader(self) -> DataLoader:
        raise NotImplementedError



import math
import numpy as np
import torch
from torch.utils.data import Dataset, DataLoader
from .loader_maker import DataLoaderBuilder


class MixtureOfGaussiansLoaderBuilder(DataLoaderBuilder):
    """DataLoader maker for mixture of Gaussians dataset."""
    def __init__(self, total_size: int, batch_size: int, num_workers=4):
        super().__init__()
        self.batch_size = batch_size
        self.num_workers = num_workers
        self.radius = 2
        self.num_centroids = 8
        self.std = 0.02

        # split into 8 : 1 : 1
        self.train_size = int(total_size * 0.8)
        self.test_size = int(total_size * 0.1)
        self.validate_size = self.test_size

    def make_test_dataloader(self):
        return self._create_dataloader(self.test_size)

    def make_train_dataloader(self):
        return self._create_dataloader(self.train_size)

    def make_validate_dataloader(self):
        return self._create_dataloader(self.validate_size)

    def _create_dataloader(self, data_size: int):
        return DataLoader(
            MixtureOfGaussiansDataset(
                self.radius, self.num_centroids, self.std, size=data_size),
            pin_memory=True,
            drop_last=True,
            num_workers=self.num_workers,
            batch_size=self.batch_size,
        )


class MixtureOfGaussiansDataset(Dataset):
    """
    Mixture of Gaussian toy dataset.
    The centroids are equally spaced around the circle of given radius.
    The centroids are chosen as uniform manner.
    """
    def __init__(self, radius: int, num_centroids: int, std: float, size: int):
        super().__init__()
        self.num_centroids = num_centroids
        self.size = size
        self.std = std

        self.diff_angle = 2 * math.pi / num_centroids
        self.center_coordinates = radius * np.array([
            [np.cos(i * self.diff_angle),
             np.sin(i * self.diff_angle)]
            for i in range(num_centroids)
        ])

    def __getitem__(self, idx):
        # uniformly choose one of the centroids
        centroid_idx = np.random.choice(self.num_centroids)
        # sample a point from a normal distribution having mean at the chosen centroid
        sampled_point = np.random.normal(
            loc=self.center_coordinates[centroid_idx], scale=self.std)
        return torch.from_numpy(sampled_point).float()

    def __len__(self):
        return self.size


if __name__ == '__main__':
    dataset = MixtureOfGaussiansDataset(
        radius=2, num_centroids=8, std=0.02, size=100)
    for i in range(10):
        print(dataset[i])



"""
VCTK clean-noisy speech dataset.

download from: https://datashare.is.ed.ac.uk/handle/10283/2791

If you unzip the file, it inflates to:
```
Archive:  DS_10283_2791.zip
  inflating: license_text
  inflating: clean_testset_wav.zip
  inflating: clean_trainset_28spk_wav.zip
  inflating: clean_trainset_56spk_wav.zip
  inflating: logfiles.zip
  inflating: noisy_testset_wav.zip
  inflating: noisy_trainset_28spk_wav.zip
  inflating: noisy_trainset_56spk_wav.zip
  inflating: testset_txt.zip
  inflating: trainset_28spk_txt.zip
  inflating: trainset_56spk_txt.zip
```
"""
import re
import os
import random
import math
import librosa
import numpy as np
from torch.utils.data import Dataset, DataLoader, sampler
from tqdm import tqdm
from torchland.utils.spectrogram import split_spectrogram, normalize_db_spectrogram
from .loader_maker import DataLoaderBuilder


N_MELS = 40  # number of mel filters
SPLIT_SIZE = 40  # chunk size to split into
# basic info of vctk dataset used for unzipping
CLEAN_TESTSET_WAV = 'clean_testset_wav'
CLEAN_TRAINSET_28SPK_WAV = 'clean_trainset_28spk_wav'
CLEAN_TRAINSET_56SPK_WAV = 'clean_trainset_56spk_wav'
NOISY_TESTSET_WAV = 'noisy_testset_wav'
NOISY_TRAINSET_28SPK_WAV = 'noisy_trainset_28spk_wav'
NOISY_TRAINSET_56SPK_WAV = 'noisy_trainset_56spk_wav'
# target directory after unzipping
CLEAN_TRAINSET_DIR = 'clean_trainset'
NOISY_TRAINSET_DIR = 'noisy_trainset'


class VCTKLoaderBuilder(DataLoaderBuilder):
    """Class that helps creating DataLoader instances for VCTK dataset."""
    def __init__(self, data_path: str, batch_size: int, num_workers=8,
                 use_channel=False, use_db_spec=False):
        super().__init__()
        self.batch_size = batch_size
        self.num_workers = num_workers

        # create datasets
        self.train_dataset = NoisyVCTKSpectrogram(
            data_path, use_channel, use_db_spec=use_db_spec)
        self.validate_dataset = NoisyVCTKSpectrogram(
            data_path, use_channel, use_db_spec=use_db_spec)
        self.test_dataset = NoisyVCTKSpectrogram(
            data_path, use_channel, use_db_spec=use_db_spec)

        # calculate indices out of dataset size
        num_data = len(self.train_dataset)
        indices = list(range(num_data))
        random.shuffle(indices)
        num_train = math.floor(num_data * 0.8)
        num_valtest = num_data - num_train
        num_validate = num_valtest // 2

        self.train_idx, valtest_idx = \
            indices[:num_train], indices[num_train:]
        self.validate_idx, self.test_idx = \
            valtest_idx[:num_validate], valtest_idx[num_validate:]

    def make_train_dataloader(self):
        train_dataloader = DataLoader(
            self.train_dataset,
            sampler=sampler.SubsetRandomSampler(self.train_idx),
            pin_memory=True,
            drop_last=True,
            num_workers=self.num_workers,
            batch_size=self.batch_size)
        return train_dataloader

    def make_validate_dataloader(self):
        validate_dataloader = DataLoader(
            self.validate_dataset,
            sampler=sampler.SubsetRandomSampler(self.validate_idx),
            pin_memory=True,
            drop_last=True,
            num_workers=self.num_workers,
            batch_size=self.batch_size)
        return validate_dataloader

    def make_test_dataloader(self):
        test_dataloader = DataLoader(
            self.test_dataset,
            sampler=sampler.SubsetRandomSampler(self.test_idx),
            pin_memory=True,
            drop_last=True,
            num_workers=self.num_workers,
            batch_size=self.batch_size)
        return test_dataloader


def unpack_dataset(fname: str, out_path: str):
    """
    Unpacks VCTK dataset.

    Args:
        fname (str): the file name of zipped vctk
        out_path (str): output directory
    """
    os.system('unzip -d {} {}'.format(out_path, fname))
    # data output directory
    clean_dir = os.path.join(out_path, CLEAN_TRAINSET_DIR)
    os.makedirs(clean_dir, exist_ok=True)
    noisy_dir = os.path.join(out_path, NOISY_TRAINSET_DIR)
    os.makedirs(noisy_dir, exist_ok=True)

    # unzip clean audio files
    os.system('unzip -d {} {}'.format(
        out_path, os.path.join(out_path, '{}.zip'.format(CLEAN_TESTSET_WAV))))
    os.system('unzip -d {} {}'.format(
        out_path, os.path.join(out_path, '{}.zip'.format(CLEAN_TRAINSET_28SPK_WAV))))
    # move the datasets for 28 and 56 speakers in the same folder
    os.system('mv {}/* {}'.format(os.path.join(out_path, CLEAN_TRAINSET_28SPK_WAV), clean_dir))
    os.system('unzip -d {} {}'.format(
        out_path, os.path.join(out_path, '{}.zip'.format(CLEAN_TRAINSET_56SPK_WAV))))
    os.system('mv {}/* {}'.format(os.path.join(out_path, CLEAN_TRAINSET_56SPK_WAV), clean_dir))

    # unzip noisy audio files
    os.system('unzip -d {} {}'.format(
        out_path, os.path.join(out_path, '{}.zip'.format(NOISY_TESTSET_WAV))))
    os.system('unzip -d {} {}'.format(
        out_path, os.path.join(out_path, '{}.zip'.format(NOISY_TRAINSET_28SPK_WAV))))
    # move the datasets for 28 and 56 speakers in the same folder
    os.system('mv {}/* {}'.format(os.path.join(out_path, NOISY_TRAINSET_28SPK_WAV), noisy_dir))
    os.system('unzip -d {} {}'.format(
        out_path, os.path.join(out_path, '{}.zip'.format(NOISY_TRAINSET_56SPK_WAV))))
    os.system('mv {}/* {}'.format(os.path.join(out_path, NOISY_TRAINSET_56SPK_WAV), noisy_dir))


def noisy_vctk_preprocess(
        in_path: str, out_path: str, clean_dir: str, noisy_dir: str,
        target_sr=16000, window_size=256, hop_size: int=None, split_size=11, mel=True):
    """
    Function for preprocessing VCTK speech data.
    The dataset should already have been unzipped to `in_path` directory.
    Preprocessed files are extracted in `.npy` format that contains
    clean, noisy spectrogram pairs.

    Each mel spectrogram has size of (n_mel, split_size).

    Args:
        in_path (str): input directory
        out_path (str): desired output directory
        clean_dir (str): clean dataset directory (input)
        noisy_dir (str): noisy dataset directory (input)
        target_sr (int): target sample rate
        window_size (int): window size
        hop_size (int): hop size
        split_size (int): number of frames of a chunk
        mel (bool): True if wanting to use mel-spectrogram
    """
    if hop_size is None:
        hop_size = window_size // 2  # 0.5 hop by default

    clean_data_path = os.path.join(in_path, clean_dir)
    print('Preprocessing : {} and {}'.format(clean_dir, noisy_dir))
    for fname in tqdm(os.listdir(clean_data_path), ascii=True):
        # parse the file name that should have form : 'p125_111.wav'
        re_match = re.match(r'p(?P<class_num>\d+)_(?P<wav_id>\d+).*\.wav', fname)
        class_num = re_match.group('class_num')
        out_dir = os.path.join(out_path, class_num)
        os.makedirs(out_dir, exist_ok=True)
        wav_id = re_match.group('wav_id')
        full_path = os.path.join(in_path, clean_dir, fname)

        # also read the noisy audio as well
        noisy_counterpart_path = os.path.join(in_path, noisy_dir, fname)

        # split spectrograms and save clean / noise pairs
        split_and_save_clean_noise_pair(
            full_path, noisy_counterpart_path,
            sr=target_sr, window_size=window_size,
            hop_size=hop_size, split_size=split_size,
            out_dir=out_dir, mel=mel)


def audiop_noisy_musicset_preprocess(
        in_path: str, out_path: str, clean_dir: str, noisy_dir: str,
        target_sr=16000, window_size=256, hop_size: int=None, split_size=11, mel=True):
    """
    Function for preprocessing audiop_noisy_musicset.
    The dataset should already have been unzipped to `in_path` directory.
    Preprocessed files are extracted in `.npy` format that contains
    clean, noisy spectrogram pairs.

    Each mel spectrogram has size of (n_mel, split_size).

    Args:
        in_path (str): input directory
        out_path (str): desired output directory
        clean_dir (str): clean dataset directory (input)
        noisy_dir (str): noisy dataset directory (input)
        target_sr (int): target sample rate
        window_size (int): window size
        hop_size (int): hop size
        split_size (int): number of frames of a chunk
        mel (bool): True if wanting to use mel-spectrogram
    """
    if hop_size is None:
        hop_size = window_size // 2  # 0.5 hop by default

    noisy_data_path = os.path.join(in_path, noisy_dir)
    print('Preprocessing : {} and {}'.format(clean_dir, noisy_dir))
    for fname in tqdm(os.listdir(noisy_data_path), ascii=True):
        # parse the file name that should have form : 
        #'<artist>_<title>_chk<chunk_num>_<noise_type>_chk<noise_chunk_num>_snr<snr>.wav'
        fname_split = fname.split('_')
        audiof_base = None
        for idx, part in enumerate(fname_split):
            if 'chk' in part:
                audiof_base = '_'.join(fname_split[:idx + 1])
                break

        noisy_full_path = os.path.join(in_path, noisy_dir, fname)

        # also read the noisy audio as well
        clean_counterpart_path = os.path.join(in_path, clean_dir, audiof_base + '.wav')

        # split spectrograms and save clean / noise pairs
        split_and_save_clean_noise_pair(
            clean_counterpart_path, noisy_full_path,
            sr=target_sr, window_size=window_size,
            hop_size=hop_size, split_size=split_size,
            out_dir=out_path, mel=mel)


def split_and_save_clean_noise_pair(
        clean_path, noisy_path, sr: int, window_size: int,
        hop_size: int, split_size: int, out_dir: str, mel=True):
    # read the audio file
    clean_y, clean_sr = librosa.load(clean_path, sr=sr)  # resample as it reads
    noisy_y, noisy_sr = librosa.load(noisy_path, sr=sr)

    assert(len(clean_y) == len(noisy_y))

    # create mel-spectrogram. perform stft otherwise
    if mel:
        # shape=(n_mels, t)
        clean_spec = librosa.feature.melspectrogram(
            clean_y, sr=sr, n_fft=window_size, hop_length=hop_size, n_mels=N_MELS)
        noisy_spec = librosa.feature.melspectrogram(
            noisy_y, sr=sr, n_fft=window_size, hop_length=hop_size, n_mels=N_MELS)
    else:
        # shape=(n_fft // 2 + 1, t)
        if clean_sr != sr:
            clean_y = librosa.core.resample(clean_y, clean_sr, sr)
        clean_spec = librosa.core.stft(clean_y, n_fft=window_size, hop_length=hop_size)
        if noisy_sr != sr:
            noisy_y = librosa.core.resample(noisy_y, clean_sr, sr)
        noisy_spec = librosa.core.stft(noisy_y, n_fft=window_size, hop_length=hop_size)

    # split the spectrogram by 'split_size'-frames-sized chunks
    clean_split = split_spectrogram(clean_spec, chunk_size_in_frames=split_size)
    noisy_split = split_spectrogram(noisy_spec, chunk_size_in_frames=split_size)

    # save each clean-noisy pairs of chunks to file
    fname = os.path.splitext(os.path.basename(clean_path))[0]
    for split_id, split_pair in enumerate(zip(clean_split, noisy_split)):
        out_fname = '{}_s{:04}'.format(fname, split_id)
        np.save(os.path.join(out_dir, out_fname),
                np.asarray(split_pair))


def noisy_custom_vctk_preprocess(
        in_path: str, out_path: str, clean_dir: str, noisy_dir: str,
        target_sr=16000, window_size=256, hop_size: int=None, split_size=11, mel=True):
    """
    TODO: temporary function for processing custom-made noisy dataset
    """
    if hop_size is None:
        hop_size = window_size // 2  # 0.5 hop by default

    clean_data_path = os.path.join(in_path, clean_dir)
    print('Preprocessing : {} and {}'.format(clean_dir, noisy_dir))

    # create data output path
    if not os.path.exists(out_path):
        os.makedirs(out_path, exist_ok=True)

    for fname in tqdm(os.listdir(clean_data_path), ascii=True):
        # parse the file name that should have form : 'p125_111.wav'
        full_path = os.path.join(in_path, clean_dir, fname)

        # also read the noisy audio as well
        noisy_counterpart_path = os.path.join(in_path, noisy_dir, fname)

        # split spectrograms and save clean / noise pairs
        split_and_save_clean_noise_pair(
            full_path, noisy_counterpart_path,
            sr=target_sr, window_size=window_size,
            hop_size=hop_size, split_size=split_size,
            out_dir=out_path, mel=mel)


class NoisyVCTKSpectrogram(Dataset):
    """
    Dataset for VCTK dataset that has been preprocessed into mel-spectrogram chunks.
    The dataset returns pairs of clean and noisy mel-spectrograms.
    """
    def __init__(self, data_path: str, use_channel=False, use_db_spec=False):
        """
        Args:
            data_path (str): data path
            use_channel (bool): add a channel dimension if True
            use_db_spec (bool): if True, data is given as db-scaled spectrogram (normalized)
        """
        super().__init__()
        self.use_channel = use_channel
        self.use_db_spec = use_db_spec

        data = []
        for root, _, files in os.walk(data_path):
            for fname in files:
                data.append(os.path.join(root, fname))
        self.data = data

    def __getitem__(self, idx):
        pair = np.load(self.data[idx])
        if self.use_channel:
            # add a channel dimension at the first (index 0) dimension
            clean = pair[0][np.newaxis, :]
            noisy = pair[1][np.newaxis, :]
        else:
            clean = pair[0]
            noisy = pair[1]

        # use db-scaled spectrum
        if self.use_db_spec:
            clean = normalize_db_spectrogram(librosa.power_to_db(clean))
            noisy = normalize_db_spectrogram(librosa.power_to_db(noisy))
        return clean, noisy

    def __len__(self):
        return len(self.data)


def preprocess():
    # run this script in order to preprocess the dataset
    path = os.path.join(os.path.dirname(__file__))

    dataset_path = os.path.join(path, 'vctk_corpus')
    unpack_dataset('DS_10283_2791.zip', out_path=dataset_path)

    out_path = 'vctk_processed'
    out_path_test = 'vctk_testset'
    # preprocess train set
    noisy_vctk_preprocess(
        in_path=dataset_path,
        out_path=out_path,
        noisy_dir=NOISY_TRAINSET_DIR,
        clean_dir=CLEAN_TRAINSET_DIR,
        split_size=SPLIT_SIZE,
    )
    # preprocess test set
    noisy_vctk_preprocess(
        in_path=dataset_path,
        out_path=out_path,
        noisy_dir=NOISY_TESTSET_WAV,
        clean_dir=CLEAN_TESTSET_WAV,
        split_size=SPLIT_SIZE,
    )

    ### UNCOMMENT BELOW TO PROCESS ORIGINAL SPECTROGRAMS FOR TRAINING SET
    # create spectrogram data also - used for waveform recovery
    # out_path_spectrogram = 'vctk_spectrogram'
    # noisy_vctk_preprocess(
    #     in_path=dataset_path,
    #     out_path=out_path_spectrogram,
    #     noisy_dir=NOISY_TRAINSET_DIR,
    #     clean_dir=CLEAN_TRAINSET_DIR,
    #     mel=False)
    noisy_vctk_preprocess(
        in_path=dataset_path,
        out_path=out_path_test,
        noisy_dir=NOISY_TESTSET_WAV,
        clean_dir=CLEAN_TESTSET_WAV,
        split_size=SPLIT_SIZE,
        mel=False,
    )




if __name__ == '__main__':
    # preprocess()
    path = os.path.join(os.path.dirname(__file__))

    dataset_path = os.path.join(path, 'audiop_noisy_musicset')

    out_path = 'audiop_noisy_musicset_processed'
    out_path_test = 'audiop_noisy_musicset_processed_test'
    # preprocess train set
    audiop_noisy_musicset_preprocess(
        in_path=dataset_path,
        out_path=out_path,
        noisy_dir='train_mixed',
        clean_dir='train_music',
        split_size=SPLIT_SIZE,
    )
    # preprocess test set
    audiop_noisy_musicset_preprocess(
        in_path=dataset_path,
        out_path=out_path,
        noisy_dir='test_mixed',
        clean_dir='test_music',
        split_size=SPLIT_SIZE,
    )

    ### UNCOMMENT BELOW TO PROCESS ORIGINAL SPECTROGRAMS FOR TRAINING SET
    # create spectrogram data also - used for waveform recovery
    # out_path_spectrogram = 'vctk_spectrogram'
    # audiop_noisy_musicset_preprocess(
    #     in_path=dataset_path,
    #     out_path=out_path_spectrogram,
    #     noisy_dir=NOISY_TRAINSET_DIR,
    #     clean_dir=CLEAN_TRAINSET_DIR,
    #     mel=False)
    audiop_noisy_musicset_preprocess(
        in_path=dataset_path,
        out_path=out_path_test,
        noisy_dir='test_mixed',
        clean_dir='test_music',
        split_size=SPLIT_SIZE,
        mel=False,
    )

    # # uncomment this for extracting local test set
    # audiop_noisy_musicset_preprocess(
    #     in_path='test',
    #     out_path='test/test_processed',
    #     noisy_dir=NOISY_TESTSET_WAV,
    #     clean_dir=CLEAN_TESTSET_WAV,
    #     split_size=SPLIT_SIZE)
    # audiop_noisy_musicset_preprocess(
    #     in_path='test',
    #     out_path='test/test_processed_spectrogram',
    #     noisy_dir=NOISY_TESTSET_WAV,
    #     clean_dir=CLEAN_TESTSET_WAV,
    #     mel=False,
    #     split_size=SPLIT_SIZE)



import pathlib
from setuptools import setup, find_packages

ROOT_PATH = pathlib.Path(__file__).parent
README_TEXT = (ROOT_PATH / 'README.md').read_text()

setup(
    name='pytorch-land',
    version='0.1.2',
    description='pytorch-land for happy deep learning',
    long_description=README_TEXT,
    long_description_content_type='text/markdown',
    author='dansuh17',
    author_email='kaehops@gmail.com',
    url='https://github.com/dansuh17/pytorch-land',
    license='MIT',
    packages=find_packages(),
)
